/* revised March & April 2021 with the following updates:
;     1.  avoid numerical underflow in GEE by setting a lower bound for regression 
;           coefficients in Poisson log-linear models and Binomial logistic models
;           (otherwise, if one class happens to have no events, then the regression
;            coefficients diverge to minus infinity and cause subsequent underflow 
;            errors)
;     2.  use extended QLIK if and only if a scale parameter is not set to 1
;     3.  stop if the relative frequency of any class is lower than a specified
;            threshold (default is 0.01, that is 1%)
;     4.  to better evaluate convergence, when considering the L2 norm for the difference between
;            the old and updated tau values, standardize per observation and class
*/

/* -------------------------------------------------------------------------------------------
;  SLTCA
;
;     Scalable and Robust Latent Trajectory Class Analysis Using Artificial Likelihood
;         Hart KR, Fei T, Hanfelt J.  Biometrics (2020)
;
;      Assumes data have been sorted by the subject id.  The calling program should specify
;         a library called RESULTS to which the output files will be written.
;
;      The calling program should also specify the following format so that SAS can properly set the reference
;         class in the polytomous logistic regression model:
;              proc format;
;                 value class_f 1='1' 2='2' 3='3' 4='4' 5='5' 6='6' 7='7' 8='8' 9='9';
;              run;
;
;
;      INPUTS:
;         runid = run id (handy way to keep track of results for multiple runs, such as when initializing 
;                         the EM algorithm using multiple random starting values)
;         seed =  starting seed for the random number generator (use 0 for computer clock)
;         num_class = number of latent classes (from 1 to 9)
;         ref = class to use as the reference when fitting a polytomous logistic regression model
;         dataset = name of data set to analyze
;         subjectid = unique id of each subject (data set must be sorted by subjectid)
;         time = time of the longitudinal observation
;         features = list of longitudinal features that are used to conceptualize the latent classes
;         dist = list of the marginal univariate distribution of each feature (currently the only
;                    allowed distributions are NOR, BIN and POI)
;         covx = list of covariates that affect the relative frequencies of the latent classes (can
;                    leave blank if desired; do not include an intercept)
;         weight = a weight to adjust for informative missing data (can leave blank if desired)
;         max_EM_iter = maximum number of EM iterations (must be at least 1)
;         initial_set = the average size of a random subset of subjects to be initially randomly assigned to
;                         latent classes (recommended:  initial_set = 25*num_class, or use 0 if the initial
;                         values of the posterior membership probabilities for all subjects are
;                         contained in a dataset called RESULTS.TAU0&runid)
;         stop_crit = 1 for stopping criterion based on L2 norm of the change in tau values (i.e., posterior
;                        membership probabilities), or 2 for stopping criterion based on the relative change 
;                        (or absolute change, whichever is more lenient) of each parameter
;         eps = tolerance of the stopping criterion
;         boundcoef = lower bound for regression coefficients in GEE Poisson log-linear models and GEE
;                      binomial logistic models
;         boundprob = lower bound for the relative frequency of any given latent class (program will stop
                       if a latent class frequency is lower than this boundary) 
;         trace = 1 to summarize the model fit of each iteration, 0 otherwise
;         show_output = 1 to show output of each GEE and LOGISTIC model fit, 0 otherwise
;         erase_log = 1 to erase the SAS log, 0 otherwise
;
;      OUTPUT FILES: (the runid will be appended to the end of these filenames)
;
;         RESULTS.TIME_FINAL    summarizes the run time, error status, and number of iterations
;         RESULTS.MODELFIT      summarizes the model fit
;         RESULTS.PARAM_FINAL   final parameter estimates
;         RESULTS.SCALE_FINAL   final scale parameter estimates
;         RESULTS.TAU_FINAL     final values of the latent class membership probabilities
;         RESULTS.TAU0          initial values of the latent class membership probabilities
;         RESULTS.MODELTRACE    (generated only if trace > 0)
;         
;    Copyright 2018, 2019, 2020 and 2021 by John Hanfelt and Kari Hart
;-------------------------------------------------------------------------------------------*/    

%macro sltca(runid=1,seed=0,num_class=2,ref=1,dataset=_LAST_,subjectid=,time=,features=,dist=,covx=,
           weight=,max_EM_iter=40,initial_set=100,boundcoef=-4,boundprob=0.01,stop_crit=1,eps=0.005,
           trace=1,show_output=0,erase_log=0);

   %IF &show_output = 0 %THEN %DO;
      ODS SELECT NONE;
   %END;

   %LET error = 0;

   data work.time;
	   now = time();
   run; 

* evaluate the number of features and distributions, and check for errors;
%local Var num_dist;
%let num_feature = 1; 
%let Var = %scan(&features, &num_feature); 
%do %while(%length(&Var));
    %let num_feature = %eval(&num_feature + 1); 
    %let Var = %scan(&features, &num_feature); 
%end; 
%let num_feature = %eval(&num_feature - 1);
%put &=num_feature;

%let num_dist = 1; 
%let Var = %scan(&dist, &num_dist); 
%do %while(%length(&Var));
   %if %upcase(%substr(&Var,1,3)) ne BIN and
       %upcase(%substr(&Var,1,3)) ne NOR and
       %upcase(%substr(&Var,1,3)) ne POI 
   %then %do;
      %put ERROR: Distribution must be BIN, NOR, or POI;
      %goto exit;
   %end;

   %let num_dist = %eval(&num_dist + 1); 
   %let Var = %scan(&dist, &num_dist); 
%end; 
%let num_dist = %eval(&num_dist - 1);
%put &=num_dist;

%if &num_feature ne &num_dist %then %do;
  %put ERROR: FEATURES and DIST vectors must have the same length;
  %goto exit;
%end;

* evaluate the number of covariates, if any; 
%let num_covx = 1; 
%let Var = %scan(&covx, &num_covx); 
%do %while(%length(&Var));
    %let num_covx = %eval(&num_covx + 1); 
    %let Var = %scan(&covx, &num_covx); 
%end; 
%let num_covx = %eval(&num_covx - 1);
%put &=num_covx;

   DATA rawdat;
	 SET &dataset; 
   RUN;

* evaluate the number of unique subjects;
     data subjects;
      set rawdat;  by &subjectid;
	  if first.&subjectid;
   run; 
   %LET num_subjects=&SYSNOBS;
   %put &=num_subjects;

   %LET subjclass = %eval(&num_class*&num_subjects);
   %put &=subjclass;

* if initial_set = 0 then the tau values have already been initialized and stored in the dataset
      RESULTS.TAU0 (with runid appended to the filename);
	
%IF &initial_set > 0 %THEN %DO;

    *** Generate initial values for tau1-tauC ;
	data results.tau0&runid;
	   
	   array tau[*] tau1 - tau&num_class;
	   SET work.rawdat (keep = &subjectid);	 by &subjectid;

       if first.&subjectid;
       call streaminit(&seed);

	   * randomly select samples to be assigned to each latent class;
	   DO g = 1 to &num_class;
	      tau[g] = 0; 
       END;

	   sample_prop = &initial_set / &num_subjects;
       u = rand("Uniform"); 
	   if u <= sample_prop then do;
 
    	   * randomly select a latent class to which the subject belongs;
           u2 = rand("Uniform");          
	       k = ceil( &num_class * u2);   
	       tau[k] = 1;
		end;
	   				
	   KEEP &subjectid tau1 - tau&num_class;
     run;

%END;

	DATA work.final;
		MERGE work.rawdat results.tau0&runid; by &subjectid;
	RUN;

		
	**************************************************************************
	*****	Prepare the WORK.MLOGIT dataset    						     *****
	*****	This is the data set used to fit the initial multinomial	 *****
	*****						logit model.							 *****
	**************************************************************************;

	*** This retains 1 observation per subject (the baseline visit).  It keeps the
		ID variable, the covariates, and the tau1-tauC values;
	DATA work.subjects;
		SET work.final;  By &subjectid;
		IF first.&subjectid;
	RUN;

    %IF &num_class = 1 %THEN %DO;

	* assign everybody to the first latent class;
      DATA work.logitpred;
		SET work.subjects (keep = &subjectid);
		IP_1 = 1;
      RUN;
	%END;
	%ELSE %DO;

 	* Each subject needs 1 observation per class;  
    * Repeat the pseudo-response data for each class, as needed for the polytomous logistic regression;
      DATA work.mlogit;

	    array tauvec[*] tau1 - tau&num_class;
		SET work.subjects;

		DO g = 1 to &num_class;
		   class = g;
           tau = tauvec[g];
		   OUTPUT;
		END;

        KEEP &subjectid class tau 
            %IF &num_covx > 0 %THEN %DO; 
                &covx 
			%END;
			;
      RUN;
	
	*** Obtain the initial estimates of alpha;

	*********************************************************
	*****	Multinomial Logit Model (0th iteration)		 *****
	*********************************************************;
	  
	   PROC LOGISTIC DATA=work.mlogit;
	      MODEL class (ref="&ref") = 
               %IF &num_covx > 0 %THEN %DO; 
                    &covx
               %END;
             / LINK=glogit;
	   	  WEIGHT tau;
	   	  OUTPUT OUT=work.logitpred predprobs=I;
	   	  ODS OUTPUT ParameterEstimates=work.logitparam
		             FitStatistics = logit_fit;
		  format class class_f.;
	   RUN;

	   data work.logitparam;
	     set work.logitparam;
	     parameter = variable;
	   * convert to numeric;
         class = 0 + response;
	     keep parameter class estimate;
	   run;

	*** Since logitpred contains IP_1-IP_C for each observation, we only 
		need to retain 1 observation per subject;
	   DATA work.logitpred;
		 SET work.logitpred;
		 BY &subjectid;
		 IF FIRST.&subjectid;

		 DROP _FROM_ _INTO_;
	   RUN;

       data logit_qic;
          set logit_fit;
          if criterion = "-2 Log L";
          qic = InterceptAndCovariates;
          keep qic;
       run;

	%END;


    *** Obtain the initial estimates of beta;
	
	*********************************************************
	*****			WGEE Model (0th iteration)			 *****
	*********************************************************;

	*** Fit the GEE model using latent class-specific weights;
    %DO j = 1 %TO &num_feature;
	    %let Var = %scan(&features, &j);
		%let Distr = %scan(&dist, &j);
        %WGEE(response = &Var, response_idx = &j, distrib = &Distr, corr=AR, tim = &time);
    %END;

    *** get cluster sizes of each feature;
	*** use the work.final dataset, not the residuals datasets, since the latter might occassionally have
	    cluster sizes of zero;
    %DO j = 1 %TO &num_feature; 
        %let Var = %scan(&features, &j);
	  	   
	       DATA work.clustsize&j;
		      retain size 0;
		      SET work.final (keep=&subjectid &Var);  by &subjectid;
              if first.&subjectid then size = 0;
	          if not(missing(&Var)) then size = size + 1;
              if last.&subjectid then output;
			  keep &subjectid size;
	          
           run;
		
     %END;


	*********************************************************
	*****  	Data set of initial parameter estimates	    *****
	*********************************************************;

	*** Create a data set containing the parameter estimates from from the multinomial logit model and 
	      the WGEE models for each feature and each class.;
	DATA work.param0;

		SET %IF &num_class > 1 %THEN %DO;
               logitparam 
			%END;
		   %DO j = 1 %TO &num_feature;
      	      %DO g = 1 %TO &num_class;
                param&g&j
              %END;
           %END;
        ;   
		estimate0 = estimate;

		KEEP parameter class feature estimate0;
	RUN;

    %LET stop = 0;
	%LET iter = 0;

	%DO %WHILE (&stop = 0);
		
		%LET iter = %eval(&iter + 1);

		TITLE;

        %update_tau; 

        %IF &num_class > 1 %THEN %DO;
               
		    DATA work.mlogit;
			   MERGE work.mlogit work.taunew;
			   tauold = tau;
			   tau = taunew;
			   drop taunew;
		    RUN;

            * new March 2021:  check for a latent class that is almost empty;
            PROC MEANS Data=work.mlogit mean;
			   CLASS class;
			   VAR tau;
               OUTPUT out=work.relfreq mean=relfreq;
		    RUN;
			data relfreq2;
			   retain myerror 0;
			   set relfreq;
               if relfreq < &boundprob then do;
			      myerror = 1;
                  CALL SYMPUTX('error', myerror); 
               end; 
		    RUN;

			%IF &stop_crit = 1 %THEN %DO;

			  *** evaluate L2 norm of the difference between TAU and TAUOLD;
              * new March 2021: standardize the L2 norm by dividing by 
			      number of observations * number of classes, then take square root;
			  DATA tau_crit;
			     retain max_tau 0;
			     set work.mlogit end=last;
				 diff = tau - tauold;
				 max_tau = max_tau + diff*diff;
				 if last then do;
                    max_tau = sqrt(max_tau/(&num_subjects. * &num_class.));
                    output;
				 end;
				 keep max_tau;
			  RUN; 			 
			   
            %END;
		%END;

		*** Rename the columns representing tau1-tauC for work.final;
		*** Assign the old values for tau1-tauC to tauold1-tauoldC;
		DATA work.final;
           array tauvec[*] tau1 - tau&num_class;
           array oldvec[*] tauold1 - tauold&num_class;

		   SET work.final;

		   DO g = 1 to &num_class;
			  oldvec[g] = tauvec[g];
		   END;
		   DROP g tau1 - tau&num_class;

		RUN;

		*** Merge the columns representing the updated tau values to
			the original data set;
        *** For this merge to work properly, it is CRUCIAL that work.final
		    does not contain variables named TAU -- otherwise, only the first
		    observation per subject would receive an updated tau value;

		DATA work.final;
			MERGE work.final work.taunew3;   BY &subjectid;
		RUN;

		%EM_iter;

        * update the latent class frequencies and QIC based on the most recent call
		  to PROC LOGISTIC;

      %IF &num_class > 1 %THEN %DO;
 
      *** Since logitpred contains tau1-tauC for each observation, we only 
			need to retain 1 observation per subject;
		DATA work.logitpred;
			SET work.logitpred;
			BY &subjectid;
			IF FIRST.&subjectid;
		RUN;

        data logit_qic;
           set logit_fit;
           if criterion = "-2 Log L";
           qic = InterceptAndCovariates;
           keep qic;
        run;

      %END;


		*********************************************************
		*****  	Data set containing parameter estimates	    *****
		*****		from kth and (k+1)th iterations			*****
		*********************************************************;

		DATA work.param;
			MERGE work.param0 work.param1;
  	        criteria = abs(estimate1 - estimate0)/max(1,abs(estimate0));
		RUN;

		PROC MEANS Data=work.param max;
			VAR criteria;
		
			*** Maximum of the relative or absolute parameter changes;
			OUTPUT out=work.param_criteria max = max_par;
		RUN;

		*** Update the macro variable stop based on stopping criteria;
		DATA _NULL_;
			stop = 0;
            %IF &num_class = 1 OR &stop_crit > 1 %THEN %DO;

			   SET work.param_criteria; 

			    *** stopping criterion for parameter estimates;  
			    IF max_par < &eps THEN stop = 1;

			%END;
			%ELSE %DO;

			    SET tau_crit;

                *** stopping criterion for posterior classification probabilities;  
			    IF max_tau < &eps THEN stop = 1;

			%END;			   

			*** stopping criterion for number of iterations;
			IF &iter >= &max_EM_iter THEN stop = 1;

			*** If an error was encountered then stop the algorithm;
			IF &error > 0 THEN stop = 1;

			*** Update the macro variable stop;
			CALL SYMPUTX('stop', stop); 
		RUN;

		*** Update the data set containing the parameter estimates;
		DATA work.param0;
			SET work.param;				
  		    estimate0 = estimate1;
			DROP estimate1;
		RUN;

		%IF &trace > 0 %THEN %DO;
		   PROC APPEND base=results.modeltrace&runid data=work.modelfit;
		   RUN;
		%END;

	%END; *** end do while loop;

	*** Save the final tau values;
	DATA results.tau_final&runid;
		SET work.taunew3;
	RUN;	
	
	DATA results.modelfit&runid;
	   SET work.modelfit;
	RUN;

    DATA results.scale_final&runid;
      SET %DO j = 1 %TO &num_feature;
      	     %DO g = 1 %TO &num_class;
                scale&g&j
              %END;
           %END;
		;
        KEEP class feature scale scale_sqrt;
     RUN;

   *** ---------------------------------------------------------- ;
   *** compute standard errors based on empirical variance matrix ;
   *** ---------------------------------------------------------- ;
    %IF &num_covx > 0 %THEN %DO; 
                    
       DATA  work.covariates; 
	       SET work.rawdat (keep=&subjectid &covx);
		   BY &subjectid;

		   IF first.&subjectid;
		   keep &covx;
	   RUN;
    %END;

    DATA work.logitpred2;
		SET work.logitpred (keep = IP_1-IP_&num_class);
	RUN;

	PROC IML;

	        * convert from a macro string variable to a string variable;
            distribs = "&dist";
			
            *** Create a matrix of cluster sizes of each feature;
			clustsize = shape(.,&num_subjects, &num_feature);

            *** Read in cluster sizes of each feature;
	        %DO j = 1 %TO &num_feature;   
	              USE work.clustsize&j;
		             read all var{size} into csize;
			      CLOSE clustsize&j;
				  clustsize[,&j] = csize;
			%END;
							
            *** read in latent class frequencies;
            USE work.logitpred2;
			   read all into pig;
		    CLOSE logitpred2;

            *** Read in scale parameters;
	        %DO j = 1 %TO &num_feature;   
                %DO g = 1 %TO &num_class;
                    USE work.scale&g&j;
						read var{scale} into scale&g&j;
					CLOSE scale&g&j;
				%END;
			%END;

			*** read in working correlation matrices;
            %DO j = 1 %TO &num_feature;   
                %DO g = 1 %TO &num_class;
                    USE work.corr&g&j;
						read all into wcorr&g&j;
					CLOSE corr&g&j;
				%END;
			%END;

		    *** read in tau values; 
            USE work.taunew2;
			   read all into tau;
		    CLOSE taunew2;

            *** read in covariates;
            %IF &num_covx > 0 %THEN %DO; 
               USE work.covariates;
		          read all into covarx;
		       CLOSE covariates;
			* include an intercept, and make sure it is in the first column;
			   nr = nrow(covarx);
			   covarx = shape(1,nr,1) || covarx;
               ncovarx = ncol(covarx);
			%END;
			%ELSE %DO;
			   covarx = 1;
			   ncovarx = 1;
			%END;
           			 
			* get the number of parameters;
           USE work.param0;
              read all var{estimate0} into estim;
           CLOSE param0;
           nparam = nrow(estim);
		 
		   bigA = shape(0,nparam,nparam);
           bigB = bigA;

		    %DO j = 1 %TO &num_feature;
                indx&j = 0;
			%END;
			
			* this loop should not be a MACRO DO loop or else might run out of symbol space;

			DO i = 1 TO &num_subjects;

   			  bigQ = shape(0,nparam,1);

		 	*** evaluate the subject- and class-specific joint estimating functions;

			  %DO g = 1 %TO &num_class;	
				qig = shape(0,nparam,1);
		
				*** score function from relative frequency model;

				%IF &num_class > 1 %THEN %DO;
				
				 DO h = 1 TO (&num_class - 1);

				    hpar = h;
					if h >= &ref then hpar = h + 1;

					DO k = 1 TO ncovarx;

					/* note the ordering of parameters: first are all the intercepts, 
					   next are params from covariate 1, etc. */
					  qindx = h + (k - 1)*(&num_class - 1);

                      %IF &num_covx > 0 %THEN %DO;
				         qig[qindx,1] = - covarx[i,k] # pig[i,hpar];
				  	     if (&g = hpar) then qig[qindx,1] = qig[qindx,1] + covarx[i,k]; 
					  %END;
					  %ELSE %DO;
					     qig[qindx,1] = - covarx # pig[i,hpar];
				  	     if (&g = hpar) then qig[qindx,1] = qig[qindx,1] + covarx; 
					  %END;
					END; 
				  END; 
				 %END;

				  %DO j = 1 %TO &num_feature;

                     size = clustsize[i,&j];
                     if size > 0 then do;
                    
				       %IF &g = 1 %THEN %DO;
                  	     indxset&j = (indx&j + 1):(indx&j + size);
				         indx&j = indx&j + size;
					   %END;

                       R_g = wcorr&g&j[1:size,1:size];

						USE work.resid&g&j;
						   read var{resid} into resid_g POINT indxset&j;
                           read var{vmu} into vmu_g POINT indxset&j;
						   read var{pred} into mu_g POINT indxset&j;
						   read var{_weight} into weight_g POINT indxset&j;
                           read var{eta_deriv} into eta_deriv_g POINT indxset&j;
					    CLOSE resid&g&j;

                        A_g = diag(vmu_g);

					*** Compute the variance-covariance matrix (feature, group, and subject specific);
					    V_g = scale&g&j * A_g##(1/2) * R_g * A_g##(1/2);
						Vinv_g = inv(V_g);
                        weight_gg = diag(weight_g);
                        Vinv_g = Vinv_g * weight_gg;
						 					
                    *** The contribution of the ith subject, gth class, jth feature to the score function;

                        /* take into account the intercept in the first column when taking the derivative 
						       of linear predictor */
                        eta_deriv_g = shape(1,size,1) || eta_deriv_g;
						nmuderiv = ncol(eta_deriv_g);

                        distrib = scan(distribs, &j); 
                        deriv_g = eta_deriv_g;
						if upcase(substr(distrib,1,3)) = "POI" then deriv_g = eta_deriv_g # mu_g;
                        if upcase(substr(distrib,1,3)) = "BIN" then deriv_g = eta_deriv_g # mu_g # (1 - mu_g);

					   	/* note the ordering of parameters:  first are all the params from feature 1, 
					       next are params from feature 2, etc. */ 						
					    qindx = ncovarx*(&num_class - 1) + nmuderiv*((&j-1)*(&num_class) + &g-1);
                        qig[(qindx + 1):(qindx+nmuderiv),1] = deriv_g` * Vinv_g * resid_g;
                       
				     END;             
                %END;    
                
				bigQ = bigQ + tau[i,&g] # qig;	
	            bigA = bigA - tau[i,&g] # (qig * qig`);
				
			%END;
			bigB = bigB + bigQ * bigQ`;

		END;

        *** now evaluate the class-specific second derivatives ;
		%DO j = 1 %TO &num_feature;
            indx&j = 0;
		%END;

		DO i = 1 TO &num_subjects;
			%DO g = 1 %TO &num_class;	
				dig = shape(0,nparam,nparam);
		
				*** second-derivatives from relative frequency model;
				%IF &num_class > 1 %THEN %DO;
				 DO h = 1 TO (&num_class - 1);
                  hpar = h;
				  if h >= &ref then hpar = h + 1;

                  DO f = 1 TO (&num_class - 1);
                   fpar = f;
				   if f >= &ref then fpar = f + 1;

					DO k = 1 TO ncovarx;
                      DO l = 1 TO ncovarx;
					  /* note the ordering of parameters: first are all the intercepts, 
					      next are params from covariate 1, etc. */
					  qindx1 = h + (k - 1)*(&num_class - 1);
                      qindx2 = f + (l - 1)*(&num_class - 1);

                       %IF &num_covx > 0 %THEN %DO;
				          dig[qindx1,qindx2] = - covarx[i,k] # covarx[i,l] # pig[i,hpar] # pig[i,fpar];
				  	      if (fpar = hpar) then dig[qindx1,qindx2] = dig[qindx1,qindx2]+covarx[i,k]#covarx[i,l]#pig[i,hpar];
					   %END;
					   %ELSE %DO;
                          dig[qindx1,qindx2] = - covarx # covarx # pig[i,hpar] # pig[i,fpar];
				  	      if (fpar = hpar) then dig[qindx1,qindx2] = dig[qindx1,qindx2]+covarx#covarx#pig[i,hpar];
					   %END;
					  END; 
					END;
				  END;
				 END; 
				%END;
				
				%DO j = 1 %TO &num_feature; 

                  size = clustsize[i,&j];
				  if size > 0 then do;
				   

                   %IF &g = 1 %THEN %DO; 
                  	  indxset&j = (indx&j + 1):(indx&j + size);
				      indx&j = indx&j + size;
				   %END;

	 				R_g = wcorr&g&j[1:size,1:size];

                       USE work.resid&g&j;
						   read var{resid} into resid_g POINT indxset&j;
                           read var{vmu} into vmu_g POINT indxset&j;
						   read var{pred} into mu_g POINT indxset&j;
						   read var{_weight} into weight_g POINT indxset&j;
                           read var{eta_deriv} into eta_deriv_g POINT indxset&j;
					    CLOSE resid&g&j;

                        A_g = diag(vmu_g);

					*** Compute the variance-covariance matrix (feature, group, and subject specific);
					    V_g = scale&g&j * A_g##(1/2) * R_g * A_g##(1/2);
						Vinv_g = inv(V_g);
                        weight_gg = diag(weight_g);
                        Vinv_g = Vinv_g * weight_gg;
						 					
                    *** The contribution of the ith subject, gth class, jth feature to the score function;

                        /* take into account the intercept in the first column when taking the 
						    derivative of linear predictor */
						eta_deriv_g = shape(1,size,1) || eta_deriv_g;
						nmuderiv = ncol(eta_deriv_g);

                        distrib = scan(distribs, &j); 
                        deriv_g = eta_deriv_g;
						if upcase(substr(distrib,1,3)) = "POI" then deriv_g = eta_deriv_g # mu_g;
                        if upcase(substr(distrib,1,3)) = "BIN" then deriv_g = eta_deriv_g # mu_g # (1 - mu_g);

					/* note the ordering of parameters:  first are all the params from feature 1, 
					     next are params from feature 2, etc. */
                    qindx = ncovarx*(&num_class - 1) + nmuderiv*((&j-1)*(&num_class) + &g-1);
                    dig[(qindx + 1):(qindx+nmuderiv),(qindx + 1):(qindx+nmuderiv)] = deriv_g` * Vinv_g * deriv_g;
                  end;                              
                %END;  

				bigA = bigA + pig[i,&g] # dig;
        	%END;               
      	END;

		bigA = bigA + bigB;
  		invbigA = inv(bigA);
		sandwich = invbigA * bigB * invbigA`;
		stderr = sqrt(vecdiag(sandwich));

        CREATE work.stderr FROM stderr[colname='se'];
		APPEND FROM stderr;
	QUIT;

	*** Save the final parameter estimates;
    %IF &syserr <= 6  %THEN %DO;
	 
	   DATA results.param_final&runid;
		   MERGE work.param0 work.stderr;
	   RUN;
	%END;
	%ELSE %DO;
	 
     * could not compute standard errors:  usually this happens when bigA is not invertible;
       DATA results.param_final&runid;
		  SET work.param0;
		  se = .;
       RUN;
	%END; 

	* when initial_set = 0, need to generate a random number in a separate data step
	    to avoid an error message that sysrandom is undefined;
    %IF &initial_set = 0 %THEN %DO;
	   data _null_;
	     call streaminit(&seed);
   	     u = rand("Uniform");
	   run;
	%END;

	%*** Create a summary of the current runid;
	data work.time2;
		set time;
		end = time();
		runtime = end - now;
		* new April 2021:  adjust runtime if the run goes past midnight;
		if runtime < 0 then runtime = (24*60*60 - now) + end;

        initseed = 0;        
		IF &initial_set > 0 THEN initseed = %eval(&sysrandom);
		iteration = %eval(&iter);
		error = &error;
		num_class = &num_class;
		KEEP num_class initseed iteration runtime error;
	run;

	proc append base=results.time_final&runid data=work.time2;
	run;
   
	%IF &erase_log > 0 %THEN %DO;
       dm "clear log";
	%END;
	%exit:  
     ;
%mend sltca;


